import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:osam/domain/event/event.dart';
import 'package:osam/domain/middleware/middleware.dart';
import 'package:osam/domain/state/base_state.dart';
import 'package:osam/persist/persist_repository.dart';

typedef void ErrorHandler(Object error);

class Store<ST extends BaseState<ST>> with PersistRepository<ST> {
  ST state;
  var middleWares = <Middleware<ST>>[];
  var _denormalizedConditions = <Condition<ST>>[];
  var logging = false;
  ErrorHandler errorHandler = (e) => print(e);

  // ignore: close_sinks
  final _dispatcher = StreamController<Event<ST>>.broadcast();

  void init() {
    _setupStoreForMiddlewares(this.middleWares);
    _denormalizeConditions();
    _dispatcher.stream.listen((event) {
      if (logging) print('Event: ${event.runtimeType}, log: ${event.log}');

      for (Condition<ST> condition in _denormalizedConditions) {
        final nextEvent = condition(event);
        if (!nextEvent) return;
      }

      try {
        event(state);
      } catch (e) {
        errorHandler(e);
      }
    });
  }

  Stream<Event<ST>> get eventStream => _dispatcher.stream;

  void dispatchEvent({@required Event<ST> event}) => _dispatcher.sink.add(event);

  void _denormalizeConditions() {
    _denormalizedConditions.clear();
    _denormalizedConditions
        .addAll(middleWares.expand((middleWare) => middleWare.conditions).toList());
  }

  void _setupStoreForMiddlewares(List<Middleware<ST>> middlewares) =>
      middlewares.forEach((middleware) => middleware.store = this);

  void updateMiddlewares(List<Middleware<ST>> middlewares) {
    _setupStoreForMiddlewares(middlewares);
    middlewares.forEach(
        (middleware) => this.middleWares[this.middleWares.indexOf(middleware)] = middleware);
    _denormalizeConditions();
  }

  void saveState() => super.save(this.state);
  void restoreState() => this.state = super.get() ?? this.state;
}
